package farkas;

import java.util.Set;
import java.util.TreeSet;
import java.util.Vector;

import DSM.DSMUtil;
import DSM.Poly;
import DSM.PolyTerm;
import net.sf.javailp.*;
import parser.Node;

public class farkasUtil {

	//infinities
	static double Inf = 1e10;
	static double negInf = -Inf;
	
	public static void addFarkasConditions(Set<Poly> preconditions, Set<Poly> postconditions) throws Exception //adds A => B using Farkas Lemma
	{
		System.err.println("Adding Farkas Condition:");
		System.err.println("========================================");
		System.err.println("Preconditions:"+preconditions);
		System.err.println("Postconditions:"+postconditions);
		System.err.println("========================================");
		
		
		//add each postcondition separately
		for(Poly post:postconditions)
			addSingleFarkasCondition(preconditions, post);
		
	}
	
	static void addSingleFarkasCondition(Set<Poly> preconditions, Poly postcondition) throws Exception
	{
		int m = preconditions.size();
		Vector<String> x = new Vector<>(); 
		x.addAll(Node.getAllPvars());
		int n = x.size();
		
		//Write the i-th precondition as A_i x <= b_i
		Vector<Vector<Double>> A = new Vector<>();
		while(A.size()<m)
			A.add(new Vector<>());
		Vector<Double> b = new Vector<>();
		int i=0;
		for(Poly prec_i:preconditions)
		{
			for(int j=0;j<n;++j)
			{
				double p = getCoefficientOfVariable(prec_i, x.elementAt(j));
				A.elementAt(i).add(p);
			}
			b.add(getConstantOf(prec_i));
			i++;
		}
		
		//write the postcondition as c^T x <= d
		Vector<Poly> c = new Vector<>();
		Poly d;
		for(int j=0;j<n;++j)
		{
			Poly p = getPolyCoefficientOfVariable(postcondition, x.elementAt(j));
			c.add(p);
		}
		d = getPolyConstantOf(postcondition);
		
		//Print the Farkas Lemma Parameters
		System.err.println("Farkas Lemma Parameters: A x <= b ==> c^T x <= d");
		System.err.println("========================================");
		System.err.println("postcondition:"+postcondition);
		System.err.println("A="+A);
		System.err.println("b="+b);
		System.err.println("c="+c);
		System.err.println("d="+d);
		System.err.println("========================================");
		
		addFarkasToLP(A, b, c, d, m, n);
	}
	
	static int ynum = 1;
	private static void addFarkasToLP(Vector<Vector<Double>> A, Vector<Double> b, Vector<Poly> c, Poly d, int m, int n) throws Exception {
		
		//create y
		Vector<String> y = new Vector<>();
		while(y.size()<m)
		{
			String yvar = "y"+ynum;
			ynum++;
			y.add(yvar);
		}
		
		//add y>=0
		for(String yvar:y)
		{
			Linear linear = new Linear();
			linear.add(1, yvar);
			problem.add(linear, ">=", 0);
			System.err.println("LP:"+linear+">=0");
		}
		
		//add A^t y = c
		for(int i=0;i<n;++i)
		{
			//c_i = \sum_{j=0}^{m-1} A_ji y_j
			//c_i - \sum_{j=0}^{m-1} A_ji y_j = 0
			Linear left = c.elementAt(i).toLinear();
			
			for(int j=0;j<m;++j)
				left.add(-A.elementAt(j).elementAt(i), y.elementAt(j));
			
			problem.add(left, "=", 0);
			System.err.println("LP:"+left+"=0");
		}
		
		//add b^t y <= d
		//equivalently d - b^t y >= 0
		Linear left = d.toLinear();
		for(int i=0;i<m;++i)
			left.add(-b.elementAt(i), y.elementAt(i));
		problem.add(left, ">=", 0);
		System.err.println("LP:"+left+">=0");
	}

	private static Poly getPolyConstantOf(Poly poly) {
		//for use in Farkas lemma, returns the constant part of poly, i.e. the part not containing any pvars
		
		Poly ans = new Poly();
		Set<String> pvars = Node.getAllPvars();
		
		for(PolyTerm t:poly.terms)
		{
			boolean addTerm = true;
			for(String pvar:pvars)
				if(t.getPowerOf(pvar)!=0)
				{
					addTerm = false;
					break;
				}
			if(addTerm)
				ans.addTerm(t.deepCopy());
		}
		
		return ans;
		
	}

	private static Poly getPolyCoefficientOfVariable(Poly poly, String var) {
		//for use in Farkas lemma
		Poly ans = new Poly();
		
		for(PolyTerm t:poly.terms)
			if(t.getPowerOf(var)==1)
			{
				PolyTerm novoT = t.deepCopy();
				novoT.powers.remove(var);
				ans.addTerm(novoT);
			}
		
		//make it negative
		ans.multiplyByConstant(-1);
		return ans;
	}

	private static double getConstantOf(Poly poly) throws Exception {//assumes poly is linear
		//for use in Farkas lemma
		//check linearity
				for(PolyTerm t:poly.terms)
					if(t.powers.size()>1)
					{
						throw new Exception("Unexpected non-linear polynomial at getConstantOf:"+poly);
					}
					else if(t.powers.size()==1 && t.getPowerOf(t.powers.firstKey())>1)
					{
						throw new Exception("Unexpected non-linear polynomial at getConstantOf:"+poly);
					}
				
		//the main part
		for(PolyTerm t:poly.terms)
			if(t.powers.size()==0)
				return t.constant;
		return 0;
	}

	private static double getCoefficientOfVariable(Poly poly, String var) throws Exception { //assumes poly is linear
		//for use in Farkas lemma
		//check linearity
		for(PolyTerm t:poly.terms)
			if(t.powers.size()>1)
			{
				throw new Exception("Unexpected non-linear polynomial at getCoefficientOfVariable:"+poly);
			}
			else if(t.powers.size()==1 && t.getPowerOf(t.powers.firstKey())>1)
			{
				throw new Exception("Unexpected non-linear polynomial at getCoefficientOfVariable:"+poly);
			}
		
		//main part
		for(PolyTerm t:poly.terms)
			if(t.getPowerOf(var)==1)
				return -t.constant;
		return 0;
	}

	public static Result solveLP() throws Exception
	{
		
		//set the objective to maximize a dummy variable that is equal to 1 -- note that oneVar is used elsewhere too and its value must be one
		Linear linear = new Linear();
		linear.add(1, "epsilon");
		problem.setObjective(linear, OptType.MAX);

		linear = new Linear();
		linear.add(1, "oneVar");

		problem.add(linear, "=", 1);
		//end of setting the objective

		//add the positivity condition for epsilon
		linear = new Linear();
		linear.add(1, "epsilon");
		problem.add(linear, ">=", 1e-9);
		
		//add freedom for all free variables
		problem.setVarLowerBound("c", negInf);
		problem.setVarUpperBound("c", Inf);
		problem.setVarLowerBound("a", negInf);
		problem.setVarUpperBound("a", Inf);
		problem.setVarLowerBound("b", negInf);
		problem.setVarUpperBound("b", Inf);
		problem.setVarLowerBound("epsilon", 0);
		problem.setVarUpperBound("epsilon", 1);
		
		//add the bounds for ys
		for(int i=1;i<ynum;++i)
		{
			problem.setVarLowerBound("y"+i, 0);
			problem.setVarUpperBound("y"+i, Inf);
		}
		
		//add the bounds for es
		for(int i=1;i<DSMUtil.eCnt;++i)
		{
			problem.setVarLowerBound("e"+i, negInf);
			problem.setVarUpperBound("e"+i, Inf);
		}
		
		
		//add a<=b
		//b-a>=0
		linear = new Linear();
		linear.add(1,"b");
		linear.add(-1,"a");
		problem.add(linear,">=",0);

		System.err.println("The LP is:");
		System.err.println("========================================");
		System.err.println(problem);
		System.err.println("========================================");
		
		
		//Create a solver and solve the LP
		Solver solver = factory.get(); 
		Result result = solver.solve(problem);

		System.err.println("The LP result is:");
		System.err.println("========================================");
		System.err.println(result);
		System.err.println("========================================");
		
		return result;
	}

	//for JavaILP
	static SolverFactory factory;
	static Problem problem;
	
	public static void initializeJavaILP() {
		factory = new SolverFactoryLpSolve(); // use lp_solve
		factory.setParameter(Solver.VERBOSE, 0); 
		factory.setParameter(Solver.TIMEOUT, 100); // set timeout to 100 seconds
		problem = new Problem();
	}
	
}
