package parser;

import java.util.HashSet;
import java.util.Set;
import java.util.TreeSet;
import java.util.Vector;

import DSM.Poly;
import DSM.PolyTerm;
import DSM.PolyUtil;

/*
 * Grammar: (rule)
 * prog := stmt ; stmt ; ... ; stmt
 * stmt := (skip) skip | (if) if bexpr then prog else prog fi | (if*) if * then prog else prog fi | (ifprob) if prob (p) then prog else prog fi | (while) while bexpr do prog od | (assignment) pvar := pexpr
 * bexpr := andbexpr or andbexpr or ... or andbexpr
 * andbexpr := literal and literal and ... and literal
 * literal := pexpr >= pexpr | pexpr <= pexpr
 * pexpr := pterm +/- pterm +/- ... +/- pterm 
 * expr := term +/- term +/- ... +/- term
 * pterm := constant * pvar
 * term := constant * var
 */

public class Node implements Comparable<Node> {
	//set of all nodes
	static Set<Node> allNodes = new HashSet<>();
	public static Vector<Node> allStatements = new Vector<>();
	
	public int statementId;
	public int id;
	public String type;
	public String rule;
	public Node parent;
	public Vector<Node> children;
	int beginIndex, endIndex;
	double value; //for constant
	public String name; //for variables
	public Node invariant; //for statements
	
	public String toString()
	{
		String ans = "Node #"+id+":\t@["+beginIndex+", "+endIndex+"]\n";
		ans += "type: "+type+"\t rule: "+rule+"\n";
		if(type.equals("constant"))
			ans+="value: "+value+"\n";
		if(type.equals("var"))
			ans+="name: "+name+"\n";
		if(type.equals("stmt"))
			ans+="statementId: "+statementId+"\n";
		if(type.equals("stmt") && invariant!=null)
			ans+="invariant: "+invariant.id+"\n";
		if(parent!=null)
			ans += "parent: "+parent.id+"\n";
		ans += "children: ";
		for(Node ch:children)
			ans+=ch.id+" ";
		ans += "\n----------------------------------------\n";
		ans = "----------------------------------------\n"+ans;
		return ans;
	}
	
	Node(Node p, String t, String r, int b, int e) throws Exception
	{
		if(b>e)
			throw new Exception("Cannot create a node with beginIndex > endIndex");
		invariant = null;
		parent = p;
		type = t;
		rule = r;
		beginIndex = b;
		endIndex = e;
		children = new Vector<>();
		//add to parent
		if(p!=null)
			p.addChild(this);
		//add to list and give id
		id = allNodes.size()+1;
		allNodes.add(this);
		//if it is a statement, add it to statements list
		if(type.equals("stmt"))
		{
			allStatements.add(this);
			statementId = allStatements.size();
		}
	}
	
	void setInvariant(Node inv) throws Exception
	{
		if(!type.equals("stmt"))
			throw new Exception("Cannot set invariant for anything other than a stmt. The node is:"+this);
		invariant = inv;
	}
	
	void setName(String n) throws Exception
	{
		if(!type.equals("var"))
			throw new Exception("Cannot set name for anything other than a var. The node is:"+this);
		name = n;
		if(name.startsWith("r") && !Parser.samplingVariablesSupport.containsKey(name))
			throw new Exception("No distribution is defined for the sampling variable "+name);
	}
	
	void setValue(double v) throws Exception
	{
		if(!type.equals("constant") )
			throw new Exception("Cannot set value for anything other than a constant. The node is:"+this);
		value = v;
	}
	
	void addChild(Node ch)
	{
		if(children.contains(ch)==false)
			children.addElement(ch);
	}
	
	public double getValue() throws Exception
	{
		if(!type.equals("constant") )
			throw new Exception("Cannot get value for anything other than a constant. The node is:"+this);
		return value;
	}
	
	String getName() throws Exception
	{
		if(!type.equals("var"))
			throw new Exception("Cannot get name for anything other than a var. The node is:"+this);
		return name;
	}
	
	public void print()
	{
		System.err.println(this);
		for(Node ch:children)
			ch.print();
	}

	public static TreeSet<String> getAllPvars() {//returns all the pvars in the program
		TreeSet<String> ans = new TreeSet<>();
		for(Node x:allNodes)
			if(x.type.equals("var") && x.rule.equals("pvar"))
				ans.add(x.name);
		return ans;
	}

	@Override
	public int compareTo(Node n) {
		if(n==null)
			return 1;
		return id-n.id;
	}

	public TreeSet<String> getRvars() { //returns all the rvars in the subtree of this node
		TreeSet<String> ans = new TreeSet<String>();
		if(rule.equals("rvar"))
			ans.add(name);
		for(Node ch:children)
		{
			TreeSet<String> chAns = ch.getRvars();
			ans.addAll(chAns);
		}
		return ans;
	}

	public Poly getPolyOfExpr() throws Exception {
		if(!type.equals("expr") && !type.equals("pexpr"))
			throw new Exception("getPolyOfExpr called for non-expr non-pexpr node:"+this);
		Poly ans = new Poly();
		
		for(Node ch:children)
		{
			PolyTerm t = ch.getPolyTermOfTerm();
			ans.addTerm(t);
		}
		
		return ans;
	}
	
	private PolyTerm getPolyTermOfTerm() throws Exception {
		if(!type.equals("pterm") && !type.equals("term"))
			throw new Exception("getPolyTermOfPterm called for non-pterm non-term node:"+this);
		PolyTerm t = new PolyTerm();
		
		for(Node ch:children)
			if(ch.type.equals("constant"))
				t.constant = ch.getValue();
			else
				t.powers.put(ch.getName(),1);
		
		return t;
	}

	public Poly getPolyOfPexpr() throws Exception
	{
		return getPolyOfExpr();
	}

	public Set<Poly> getPolySetOfBexpr() throws Exception {
		if(!type.equals("bexpr"))
			throw new Exception("getPolySetOfBexpr called for non-bexpr node:"+this);
		if(children.size()==0)
		{
			//return 1>=0
			Set<Poly> ans = new HashSet<Poly>();
			PolyTerm t = new PolyTerm();
			t.constant=1;
			Poly p = new Poly();
			p.addTerm(t);
			ans.add(p);
			return ans;
		}
		else if(children.size()==1)
			return children.firstElement().getPolySetOfAndbexpr();
		else
			throw new Exception("getPolySetOfBexpr cannot handle the or operation.");
	}
	
	public Set<Poly> getPolySetOfAndbexpr() throws Exception
	{
		if(!type.equals("andbexpr"))
			throw new Exception("getPolySetOfAndbexpr called for non-andbexpr node:"+this);
		Set<Poly> ans = new HashSet<Poly>();
		
		for(Node literal:children)
			ans.add(literal.getPolyOfLiteral());
		
		if(ans.size()==0) //add 1>=0
		{
			PolyTerm t = new PolyTerm();
			t.constant=1;
			Poly p = new Poly();
			p.addTerm(t);
			ans.add(p);
		}
		
		return ans;
	}
	
	public Poly getPolyOfLiteral() throws Exception
	{
		if(!type.equals("literal"))
			throw new Exception("getPolyOfLiteral called for non-literal node:"+this);
		
		Poly bigger, smaller;
		if(rule.equals("<="))
		{
			smaller = children.elementAt(0).getPolyOfPexpr();
			bigger = children.elementAt(1).getPolyOfPexpr();
		}
		else
		{
			smaller = children.elementAt(1).getPolyOfPexpr();
			bigger = children.elementAt(0).getPolyOfPexpr();
		}
		
		//bigger-smaller>=0
		Poly ans = PolyUtil.subtract(bigger, smaller);
		
		return ans;
	}
	
}
